@@ -1,7 +1,7 @@
-"""
-The PythonInfo contains information about a concrete instance of a Python interpreter.
+"""The PythonInfo contains information about a concrete instance of a Python interpreter.
 
 Note: this file is also used to query target interpreters, so can only use standard library methods
+
 """
 
 from __future__ import annotations
@@ -17,6 +17,14 @@ import sysconfig
 import warnings
 from collections import OrderedDict, namedtuple
 from string import digits
+from typing import TYPE_CHECKING
+
+if TYPE_CHECKING:
+    from collections.abc import Mapping
+
+    from virtualenv.app_data.base import AppData
+    from virtualenv.discovery.py_spec import PythonSpec
+    from virtualenv.run.plugin.creators import CreatorSelector
 
 VersionInfo = namedtuple("VersionInfo", ["major", "minor", "micro", "releaselevel", "serial"])  # noqa: PYI024
 LOGGER = logging.getLogger(__name__)
@@ -27,10 +35,10 @@ def _get_path_extensions():
 
 
 EXTENSIONS = _get_path_extensions()
-_CONF_VAR_RE = re.compile(r"\{\w+\}")
+_CONF_VAR_RE = re.compile(r"\{\w+}")
 
 
-class PythonInfo:
+class PythonInfo:  # noqa: PLR0904
     """Contains information for a Python interpreter."""
 
     def __init__(self) -> None:  # noqa: PLR0915
@@ -41,7 +49,7 @@ class PythonInfo:
         self.platform = sys.platform
         self.implementation = platform.python_implementation()
         if self.implementation == "PyPy":
-            self.pypy_version_info = tuple(sys.pypy_version_info)
+            self.pypy_version_info = tuple(sys.pypy_version_info)  # ty: ignore[unresolved-attribute]
 
         # this is a tuple in earlier, struct later, unify to our own named tuple
         self.version_info = VersionInfo(*sys.version_info)
@@ -50,6 +58,11 @@ class PythonInfo:
         # pointer size.
         self.architecture = 32 if struct.calcsize("P") == 4 else 64  # noqa: PLR2004
 
+        # Full platform identity string from Python's build configuration.
+        # Used to derive the instruction set architecture (ISA) via the `machine` property.
+        # sysconfig.get_platform() returns e.g. 'macosx-14.0-arm64', 'linux-x86_64', 'win-amd64', 'win32'.
+        self.sysconfig_platform = sysconfig.get_platform()
+
         # Used to determine some file names.
         # See `CPython3Windows.python_zip()`.
         self.version_nodot = sysconfig.get_config_var("py_version_nodot")
@@ -111,7 +124,7 @@ class PythonInfo:
             k: v
             for k, v in [
                 # a list of content to store from sysconfig
-                ("makefile_filename", makefile()),
+                ("makefile_filename", makefile() if makefile is not None else None),
             ]
             if k is not None
         }
@@ -120,6 +133,7 @@ class PythonInfo:
         for element in self.sysconfig_paths.values():
             config_var_keys.update(k[1:-1] for k in _CONF_VAR_RE.findall(element))
         config_var_keys.add("PYTHONFRAMEWORK")
+        config_var_keys.update(("Py_ENABLE_SHARED", "INSTSONAME", "LIBDIR"))
 
         self.sysconfig_vars = {i: sysconfig.get_config_var(i or "") for i in config_var_keys}
 
@@ -129,19 +143,20 @@ class PythonInfo:
             self.tcl_lib, self.tk_lib = None, None
 
         confs = {
-            k: (self.system_prefix if v is not None and v.startswith(self.prefix) else v)
+            k: (self.system_prefix if isinstance(v, str) and v.startswith(self.prefix) else v)
             for k, v in self.sysconfig_vars.items()
         }
         self.system_stdlib = self.sysconfig_path("stdlib", confs)
         self.system_stdlib_platform = self.sysconfig_path("platstdlib", confs)
         self.max_size = getattr(sys, "maxsize", getattr(sys, "maxint", None))
+        self._creators = None
 
     @staticmethod
     def _get_tcl_tk_libs():
-        """
-        Detects the tcl and tk libraries using tkinter.
+        """Detects the tcl and tk libraries using tkinter.
 
         This works reliably but spins up tkinter, which is heavy if you don't need it.
+
         """
         tcl_lib, tk_lib = None, None
         try:
@@ -216,15 +231,15 @@ class PythonInfo:
         return self._try_posix_fallback_executable(base_executable)
 
     def _try_posix_fallback_executable(self, base_executable):
-        """
-        Try to find a versioned Python binary as fallback for POSIX virtual environments.
+        """Try to find a versioned Python binary as fallback for POSIX virtual environments.
 
-        Python may return "python" because it was invoked from the POSIX virtual environment
-        however some installs/distributions do not provide a version-less "python" binary in
-        the system install location (see PEP 394) so try to fallback to a versioned binary.
+        Python may return "python" because it was invoked from the POSIX virtual environment however some
+        installs/distributions do not provide a version-less "python" binary in the system install location (see PEP
+        394) so try to fallback to a versioned binary.
+
+        Gate this to Python 3.11 as `sys._base_executable` path resolution is now relative to the 'home' key from
+        pyvenv.cfg which often points to the system install location.
 
-        Gate this to Python 3.11 as `sys._base_executable` path resolution is now relative to
-        the 'home' key from pyvenv.cfg which often points to the system install location.
         """
         major, minor = self.version_info.major, self.version_info.minor
         if self.os != "posix" or (major, minor) < (3, 11):
@@ -235,6 +250,8 @@ class PythonInfo:
         candidates = [f"python{major}", f"python{major}.{minor}"]
         if self.implementation == "PyPy":
             candidates.extend(["pypy", "pypy3", f"pypy{major}", f"pypy{major}.{minor}"])
+        elif self.implementation == "RustPython":
+            candidates.extend(["rustpython", f"rustpython{major}", f"rustpython{major}.{minor}"])
 
         for candidate in candidates:
             full_path = os.path.join(base_dir, candidate)
@@ -243,7 +260,14 @@ class PythonInfo:
 
         return None  # in this case we just can't tell easily without poking around FS and calling them, bail
 
-    def install_path(self, key):
+    def install_path(self, key: str) -> str:
+        """Return the relative installation path for a given installation scheme key.
+
+        :param key: the sysconfig scheme key (e.g. ``purelib``, ``headers``)
+
+        :returns: the relative path for the given key
+
+        """
         result = self.distutils_install.get(key)
         if result is None:  # use sysconfig if sysconfig_scheme is set or distutils is unavailable
             # set prefixes to empty => result is relative from cwd
@@ -260,8 +284,12 @@ class PythonInfo:
         with warnings.catch_warnings():  # disable warning for PEP-632
             warnings.simplefilter("ignore")
             try:
-                from distutils import dist  # noqa: PLC0415
-                from distutils.command.install import SCHEME_KEYS  # noqa: PLC0415
+                from distutils import (  # noqa: PLC0415  # ty: ignore[unresolved-import]  # https://github.com/astral-sh/ty/issues/708
+                    dist,
+                )
+                from distutils.command.install import (  # noqa: PLC0415  # ty: ignore[unresolved-import]  # https://github.com/astral-sh/ty/issues/708
+                    SCHEME_KEYS,
+                )
             except ImportError:  # if removed or not installed ignore
                 return {}
 
@@ -278,27 +306,41 @@ class PythonInfo:
         return {key: (getattr(i, f"install_{key}")[1:]).lstrip(os.sep) for key in SCHEME_KEYS}
 
     @property
-    def version_str(self):
+    def version_str(self) -> str:
+        """The full version as ``major.minor.micro`` string (e.g. ``3.13.2``)."""
         return ".".join(str(i) for i in self.version_info[0:3])
 
     @property
-    def version_release_str(self):
+    def version_release_str(self) -> str:
+        """The release version as ``major.minor`` string (e.g. ``3.13``)."""
         return ".".join(str(i) for i in self.version_info[0:2])
 
     @property
-    def python_name(self):
+    def python_name(self) -> str:
+        """The python executable name as ``pythonX.Y`` (e.g. ``python3.13``)."""
         version_info = self.version_info
         return f"python{version_info.major}.{version_info.minor}"
 
     @property
-    def is_old_virtualenv(self):
+    def is_old_virtualenv(self) -> bool:
+        """``True`` if this interpreter runs inside an old-style virtualenv (has ``real_prefix``)."""
         return self.real_prefix is not None
 
     @property
-    def is_venv(self):
+    def is_venv(self) -> bool:
+        """``True`` if this interpreter runs inside a PEP 405 venv (has ``base_prefix``)."""
         return self.base_prefix is not None
 
-    def sysconfig_path(self, key, config_var=None, sep=os.sep):
+    def sysconfig_path(self, key: str, config_var: dict[str, str] | None = None, sep: str = os.sep) -> str:
+        """Return the sysconfig install path for a scheme key, optionally substituting config variables.
+
+        :param key: the sysconfig path name (e.g. ``stdlib``, ``purelib``)
+        :param config_var: override config variables for path expansion, or ``None`` to use defaults
+        :param sep: path separator to use
+
+        :returns: the expanded path string
+
+        """
         pattern = self.sysconfig_paths.get(key)
         if pattern is None:
             return ""
@@ -310,12 +352,27 @@ class PythonInfo:
             config_var = base
         return pattern.format(**config_var).replace("/", sep)
 
+    def creators(self, refresh: bool = False) -> CreatorSelector:  # noqa: FBT001, FBT002
+        """Return the available creators for this interpreter, cached unless ``refresh`` is ``True``.
+
+        :param refresh: if ``True``, re-discover creators instead of using the cache
+
+        :returns: the creator selector for this interpreter
+
+        """
+        if self._creators is None or refresh is True:
+            from virtualenv.run.plugin.creators import CreatorSelector  # noqa: PLC0415
+
+            self._creators = CreatorSelector.for_interpreter(self)
+        return self._creators
+
     @property
-    def system_include(self):
+    def system_include(self) -> str:
+        """The path to the system include directory for C headers."""
         path = self.sysconfig_path(
             "include",
             {
-                k: (self.system_prefix if v is not None and v.startswith(self.prefix) else v)
+                k: (self.system_prefix if isinstance(v, str) and v.startswith(self.prefix) else v)
                 for k, v in self.sysconfig_vars.items()
             },
         )
@@ -327,11 +384,13 @@ class PythonInfo:
         return path
 
     @property
-    def system_prefix(self):
+    def system_prefix(self) -> str:
+        """The prefix of the system Python this interpreter is based on."""
         return self.real_prefix or self.base_prefix or self.prefix
 
     @property
-    def system_exec_prefix(self):
+    def system_exec_prefix(self) -> str:
+        """The exec prefix of the system Python this interpreter is based on."""
         return self.real_prefix or self.base_exec_prefix or self.exec_prefix
 
     def __repr__(self) -> str:
@@ -369,24 +428,59 @@ class PythonInfo:
         )
 
     @property
-    def spec(self):
-        return "{}{}{}-{}".format(
+    def machine(self) -> str:
+        """The instruction set architecture (ISA) of this interpreter.
+
+        Derived from :func:`sysconfig.get_platform`, which returns the platform identity string used by Python's build
+        system (e.g. ``macosx-14.0-arm64``, ``linux-x86_64``, ``win-amd64``). The ISA is the last component of this
+        string.
+
+        For macOS ``universal2`` fat binaries, falls back to :func:`platform.machine` to report the actual runtime ISA
+        (``arm64`` or ``x86_64``) rather than the build target.
+
+        :returns: the ISA string, e.g. ``arm64``, ``x86_64``, ``x86``
+
+        """
+        if (plat := self.sysconfig_platform) is None:
+            return "unknown"
+        if plat == "win32":
+            return "x86"
+        if (isa := plat.rsplit("-", 1)[-1]) == "universal2":
+            isa = platform.machine().lower()
+        return normalize_isa(isa)
+
+    @property
+    def spec(self) -> str:
+        """A specification string identifying this interpreter (e.g. ``CPython3.13.2-64-arm64``)."""
+        return "{}{}{}-{}-{}".format(
             self.implementation,
             ".".join(str(i) for i in self.version_info),
             "t" if self.free_threaded else "",
             self.architecture,
+            self.machine,
         )
 
     @classmethod
-    def clear_cache(cls, cache=None):
-        # this method is not used by itself, so here and called functions can import stuff locally
+    def clear_cache(cls, app_data: AppData) -> None:
+        """Clear all cached interpreter information from the application data store.
+
+        :param app_data: the application data store to clear
+
+        """
         from virtualenv.discovery.cached_py_info import clear  # noqa: PLC0415
 
-        clear(cache)
+        clear(app_data)
         cls._cache_exe_discovery.clear()
 
-    def satisfies(self, spec, impl_must_match):  # noqa: C901, PLR0911
-        """Check if a given specification can be satisfied by the this python interpreter instance."""
+    def satisfies(self, spec: PythonSpec, impl_must_match: bool) -> bool:  # noqa: C901, FBT001, PLR0911, PLR0912
+        """Check if a given specification can be satisfied by this python interpreter instance.
+
+        :param spec: the python specification to check against
+        :param impl_must_match: if ``True``, the implementation (e.g. CPython vs PyPy) must also match
+
+        :returns: ``True`` if this interpreter satisfies the specification
+
+        """
         if spec.path:
             if self.executable == os.path.abspath(spec.path):
                 return True  # if the path is a our own executable path we're done
@@ -411,9 +505,26 @@ class PythonInfo:
         if spec.architecture is not None and spec.architecture != self.architecture:
             return False
 
+        if spec.machine is not None and spec.machine != self.machine:
+            return False
+
         if spec.free_threaded is not None and spec.free_threaded != self.free_threaded:
             return False
 
+        if spec.version_specifier is not None:
+            version_info = self.version_info
+            release = f"{version_info.major}.{version_info.minor}.{version_info.micro}"
+            if version_info.releaselevel != "final":
+                suffix = {
+                    "alpha": "a",
+                    "beta": "b",
+                    "candidate": "rc",
+                }.get(version_info.releaselevel)
+                if suffix is not None:
+                    release = f"{release}{suffix}{version_info.serial}"
+            if not spec.version_specifier.contains(release):
+                return False
+
         for our, req in zip(self.version_info[0:3], (spec.major, spec.minor, spec.micro)):
             if req is not None and our is not None and our != req:
                 return False
@@ -423,35 +534,35 @@ class PythonInfo:
     _current = None
 
     @classmethod
-    def current(cls, app_data, cache):
+    def current(cls, app_data: AppData | None = None) -> PythonInfo:
+        """Locate the current host interpreter information.
+
+        This might differ from the running interpreter if the host Python has been upgraded.
+
+        :param app_data: the application data store for caching
+
+        :returns: the current interpreter information
+
         """
-        This locates the current host interpreter information. This might be different than what we run into in case
-        the host python has been upgraded from underneath us.
-        """  # noqa: D205
         if cls._current is None:
-            cls._current = cls.from_exe(
-                sys.executable,
-                app_data,
-                cache,
-                raise_on_error=True,
-                resolve_to_host=False,
-            )
+            result = cls.from_exe(sys.executable, app_data, raise_on_error=True, resolve_to_host=False)
+            assert result is not None  # raise_on_error=True guarantees non-None  # noqa: S101
+            cls._current = result
         return cls._current
 
     @classmethod
-    def current_system(cls, app_data, cache) -> PythonInfo:
+    def current_system(cls, app_data: AppData | None = None) -> PythonInfo:
+        """Locate the current system interpreter information, resolving through any virtualenv layers.
+
+        :param app_data: the application data store for caching
+
+        :returns: the system interpreter information
+
         """
-        This locates the current host interpreter information. This might be different than what we run into in case
-        the host python has been upgraded from underneath us.
-        """  # noqa: D205
         if cls._current_system is None:
-            cls._current_system = cls.from_exe(
-                sys.executable,
-                app_data,
-                cache,
-                raise_on_error=True,
-                resolve_to_host=True,
-            )
+            result = cls.from_exe(sys.executable, app_data, raise_on_error=True, resolve_to_host=True)
+            assert result is not None  # raise_on_error=True guarantees non-None  # noqa: S101
+            cls._current_system = result
         return cls._current_system
 
     def _to_json(self):
@@ -459,39 +570,43 @@ class PythonInfo:
         return json.dumps(self._to_dict(), indent=2)
 
     def _to_dict(self):
-        data = {var: getattr(self, var) for var in vars(self)}
-        data["version_info"] = data["version_info"]._asdict()  # namedtuple to dictionary
+        data = {var: (getattr(self, var) if var != "_creators" else None) for var in vars(self)}
+
+        version_info = data["version_info"]
+        data["version_info"] = version_info._asdict() if hasattr(version_info, "_asdict") else version_info
         return data
 
     @classmethod
     def from_exe(  # noqa: PLR0913
         cls,
-        exe,
-        app_data,
-        cache,
-        raise_on_error=True,  # noqa: FBT002
-        ignore_cache=False,  # noqa: FBT002
-        resolve_to_host=True,  # noqa: FBT002
-        env=None,
-    ):
-        """Given a path to an executable get the python information."""
+        exe: str,
+        app_data: AppData | None = None,
+        raise_on_error: bool = True,  # noqa: FBT001, FBT002
+        ignore_cache: bool = False,  # noqa: FBT001, FBT002
+        resolve_to_host: bool = True,  # noqa: FBT001, FBT002
+        env: Mapping[str, str] | None = None,
+    ) -> PythonInfo | None:
+        """Given a path to an executable, get the python information.
+
+        :param exe: path to the Python executable
+        :param app_data: the application data store for caching
+        :param raise_on_error: if ``True``, raise on failure instead of returning ``None``
+        :param ignore_cache: if ``True``, bypass the cache and re-query the interpreter
+        :param resolve_to_host: if ``True``, resolve through virtualenv layers to the host interpreter
+        :param env: environment variables to use for subprocess calls
+
+        :returns: the interpreter information, or ``None`` if discovery failed and ``raise_on_error`` is ``False``
+
+        """
         # this method is not used by itself, so here and called functions can import stuff locally
-        from virtualenv.discovery.cached_py_info import from_exe as from_exe_cache  # noqa: PLC0415
+        from virtualenv.discovery.cached_py_info import from_exe  # noqa: PLC0415
 
         env = os.environ if env is None else env
-        proposed = from_exe_cache(
-            cls,
-            app_data,
-            exe,
-            env=env,
-            raise_on_error=raise_on_error,
-            ignore_cache=ignore_cache,
-            cache=cache,
-        )
+        proposed = from_exe(cls, app_data, exe, env=env, raise_on_error=raise_on_error, ignore_cache=ignore_cache)
 
         if isinstance(proposed, PythonInfo) and resolve_to_host:
             try:
-                proposed = proposed._resolve_to_system(app_data, proposed, cache=cache)  # noqa: SLF001
+                proposed = proposed._resolve_to_system(app_data, proposed)  # noqa: SLF001
             except Exception as exception:
                 if raise_on_error:
                     raise
@@ -513,7 +628,7 @@ class PythonInfo:
         return result
 
     @classmethod
-    def _resolve_to_system(cls, app_data, target, cache):
+    def _resolve_to_system(cls, app_data, target):
         start_executable = target.executable
         prefixes = OrderedDict()
         while target.system_executable is None:
@@ -530,15 +645,33 @@ class PythonInfo:
                 msg = "prefixes are causing a circle {}".format("|".join(prefixes.keys()))
                 raise RuntimeError(msg)
             prefixes[prefix] = target
-            target = target.discover_exe(app_data, prefix=prefix, exact=False, cache=cache)
+            target = target.discover_exe(app_data, prefix=prefix, exact=False)
         if target.executable != target.system_executable:
-            target = cls.from_exe(target.system_executable, app_data, cache)
+            resolved = cls.from_exe(target.system_executable, app_data)
+            if resolved is not None:
+                target = resolved
         target.executable = start_executable
         return target
 
     _cache_exe_discovery = {}  # noqa: RUF012
 
-    def discover_exe(self, app_data, cache, prefix, exact=True, env=None):  # noqa: FBT002
+    def discover_exe(
+        self,
+        app_data: AppData,
+        prefix: str,
+        exact: bool = True,  # noqa: FBT001, FBT002
+        env: Mapping[str, str] | None = None,
+    ) -> PythonInfo:
+        """Discover a matching Python executable under a given prefix directory.
+
+        :param app_data: the application data store for caching
+        :param prefix: the directory prefix to search under
+        :param exact: if ``True``, require an exact version match; if ``False``, pick the closest match
+        :param env: environment variables to use for subprocess calls
+
+        :returns: the discovered interpreter information
+
+        """
         key = prefix, exact
         if key in self._cache_exe_discovery and prefix:
             LOGGER.debug("discover exe from cache %s - exact %s: %r", prefix, exact, self._cache_exe_discovery[key])
@@ -551,7 +684,7 @@ class PythonInfo:
         env = os.environ if env is None else env
         for folder in possible_folders:
             for name in possible_names:
-                info = self._check_exe(app_data, cache, folder, name, exact, discovered, env)
+                info = self._check_exe(app_data, folder, name, exact, discovered, env)
                 if info is not None:
                     self._cache_exe_discovery[key] = info
                     return info
@@ -564,21 +697,14 @@ class PythonInfo:
         msg = "failed to detect {} in {}".format("|".join(possible_names), os.pathsep.join(possible_folders))
         raise RuntimeError(msg)
 
-    def _check_exe(self, app_data, cache, folder, name, exact, discovered, env):  # noqa: PLR0913
+    def _check_exe(self, app_data, folder, name, exact, discovered, env):  # noqa: PLR0913
         exe_path = os.path.join(folder, name)
         if not os.path.exists(exe_path):
             return None
-        info = self.from_exe(
-            exe_path,
-            app_data,
-            cache,
-            resolve_to_host=False,
-            raise_on_error=False,
-            env=env,
-        )
+        info = self.from_exe(exe_path, app_data, resolve_to_host=False, raise_on_error=False, env=env)
         if info is None:  # ignore if for some reason we can't query
             return None
-        for item in ["implementation", "architecture", "version_info"]:
+        for item in ["implementation", "architecture", "machine", "version_info"]:
             found = getattr(info, item)
             searched = getattr(self, item)
             if found != searched:
@@ -599,12 +725,13 @@ class PythonInfo:
         # could cause this (when using copy strategy of the host python)
         def sort_by(info):
             # we need to setup some priority of traits, this is as follows:
-            # implementation, major, minor, micro, architecture, tag, serial
+            # implementation, major, minor, architecture, machine, micro, tag, serial
             matches = [
                 info.implementation == target.implementation,
                 info.version_info.major == target.version_info.major,
                 info.version_info.minor == target.version_info.minor,
                 info.architecture == target.architecture,
+                info.machine == target.machine,
                 info.version_info.micro == target.version_info.micro,
                 info.version_info.releaselevel == target.version_info.releaselevel,
                 info.version_info.serial == target.version_info.serial,
@@ -659,8 +786,7 @@ class PythonInfo:
         for base in possible_base:
             lower = base.lower()
             yield lower
-
-            from virtualenv.discovery.info import fs_is_case_sensitive  # noqa: PLC0415
+            from virtualenv.info import fs_is_case_sensitive  # noqa: PLC0415
 
             if fs_is_case_sensitive():
                 if base != lower:
@@ -670,6 +796,10 @@ class PythonInfo:
                     yield upper
 
 
+def normalize_isa(isa: str) -> str:
+    return {"amd64": "x86_64", "aarch64": "arm64"}.get(low := isa.lower(), low)
+
+
 if __name__ == "__main__":
     # dump a JSON representation of the current python
 
@@ -691,3 +821,4 @@ if __name__ == "__main__":
 
     info = PythonInfo()._to_json()  # noqa: SLF001
     sys.stdout.write("".join((start_cookie[::-1], info, end_cookie[::-1])))
+    sys.stdout.flush()
