@@ -6,7 +6,6 @@ import subprocess
 import sys
 from os.path import dirname, normcase
 from pathlib import Path
-from shlex import quote
 from subprocess import Popen
 
 import pytest
@@ -77,13 +76,18 @@ class ActivationTester:
 
         try:
             process = Popen(invoke, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)
-            raw_, _ = process.communicate()
-            raw = raw_.decode()
-            assert process.returncode == 0, raw
+            raw_, _ = process.communicate(timeout=120)
+        except subprocess.TimeoutExpired:
+            process.kill()
+            process.communicate()
+            pytest.fail(f"Activation script timed out: {invoke}")
         except subprocess.CalledProcessError as exception:
             output = exception.output + exception.stderr
             assert not exception.returncode, output  # noqa: PT017
             return None
+        else:
+            raw = raw_.decode(errors="replace")
+            assert process.returncode == 0, raw
 
         out = re.sub(r"pydev debugger: process \d+ is connecting\n\n", "", raw, flags=re.MULTILINE).strip().splitlines()
         self.assert_output(out, raw, tmp_path)
@@ -131,13 +135,13 @@ class ActivationTester:
         ]
 
     def assert_output(self, out, raw, tmp_path):
-        # pre-activation
+        """Compare _get_test_lines() with the expected values."""
         assert out[0], raw
         assert out[1] == "None", raw
         assert out[2] == "None", raw
-        # post-activation
-        expected = self._creator.exe.parent / os.path.basename(sys.executable)
-        assert self.norm_path(out[3]) == self.norm_path(expected), raw
+        # self.activate_call(activate_script) runs at this point
+        python_exe = self._creator.exe.parent / os.path.basename(sys.executable)
+        assert self.norm_path(out[3]) == self.norm_path(python_exe), raw
         assert self.norm_path(out[4]) == self.norm_path(self._creator.dest).replace("\\\\", "\\"), raw
         assert out[5] == self._creator.env_name
         # Some attempts to test the prompt output print more than 1 line.
@@ -154,7 +158,7 @@ class ActivationTester:
         assert out[-1] == "None", raw
 
     def quote(self, s):
-        return quote(s)
+        return self.of_class.quote(s)
 
     def python_cmd(self, cmd):
         return f"{os.path.basename(sys.executable)} -c {self.quote(cmd)}"
@@ -213,8 +217,8 @@ class RaiseOnNonSourceCall(ActivationTester):
             stderr=subprocess.PIPE,
             env=env,
         )
-        _out, _err = process.communicate()
-        err = _err.decode("utf-8")
+        _out, err_ = process.communicate()
+        err = err_.decode("utf-8")
         assert process.returncode
         assert self.non_source_fail_message in err
 
@@ -233,6 +237,7 @@ def raise_on_non_source_class():
 def activation_python(request, tmp_path_factory, special_char_name, current_fastest):
     dest = os.path.join(str(tmp_path_factory.mktemp("activation-tester-env")), special_char_name)
     cmd = ["--without-pip", dest, "--creator", current_fastest, "-vv", "--no-periodic-update"]
+    # `params` is accessed here. https://docs.pytest.org/en/stable/reference/reference.html#pytest-fixture
     if request.param:
         cmd += ["--prompt", special_char_name]
     session = cli_run(cmd)
